(*
  This program allows a lot of benchs
  
*)

module Misc =
struct

  let pseudo_random_int i = Random.int i

  (* safe_print_string : string -> unit *)
  (** locks a global mutex, prints its given string, unlocks the mutex *)
  let safe_print_string =
    let mutex_print = Mutex.create () in
      (fun s ->
	 Mutex.lock mutex_print ;
	 print_string s ;
	 flush stdout ;
	 Mutex.unlock mutex_print)

  let safe_prerr_string =
    let mutex_print = Mutex.create () in
      (fun s ->
	 Mutex.lock mutex_print ;
	 prerr_string s ;
	 flush stderr ;
	 Mutex.unlock mutex_print)

  (* intervalle : int -> int list *)
  (** [intervalle a b] returns the list of the numbers between [a] and [b]
      // tail-recursive *)
  let intervalle = 
    let rec intervalle r a b =
      if a > b then r
      else intervalle (a::r) (a+1) b
    in intervalle []
end

module Trace =
struct
  let mode = ref false
  let dummy =
    mode :=
      try 
	if Sys.getenv "TRACE" = "ON" then
	  (Misc.safe_prerr_string "TRACE ON! (export TRACE=OFF to disactivate)\n"; true)
	else
	  (Misc.safe_prerr_string "TRACE OFF! (export TRACE=ON to activate)\n"; false)
      with _ ->
	(Misc.safe_prerr_string "TRACE OFF! (export TRACE=ON to activate)\n"; false)
  let safe_print_string s = if !mode then Misc.safe_prerr_string s else ()
end


module type Unit = sig end
module Unit : Unit = struct end
module Main = functor (M : sig val launch : unit -> unit end) ->
struct
  let main numberofthreads =
    Array.iteri
      (fun i e ->
	 Thread.join e;
	 Trace.safe_print_string (Printf.sprintf  "Thread #%d joined.\n" i))
      (Array.mapi
	 (fun i () ->
	    let t = Thread.create M.launch () in
	      Trace.safe_print_string (Printf.sprintf  "Thread #%d created.\n" i);
	      t)
	 (Array.make numberofthreads ()));
    print_endline "I've waited for the main threads once.";
end

module Intervalle = functor (Dummy : Unit) ->
struct
  (* let these mutual rec functions be at the top level of this module *)
  let rec intervalle r a b =
  if a > b then (Trace.safe_print_string "intervalle"; r)
  else intervalle2 (a::r) (a+1) b

  and intervalle2 r a b =
  if a > b then []
  else intervalle (a::r) (a+1) b

  let intervalle a b = intervalle [] a b

  include Main(struct let launch () = ignore(intervalle 0 100_000) end)
end

module Loops = functor (Dummy : Unit) ->
struct
  let loop () = 
    for i = 0 to max_int do
      ()
    done

  include Main(struct let launch () = ignore(loop ()) end)
end

module LoopsAlloc = functor (Dummy : Unit) ->
struct
  let loop () = 
    for i = 0 to max_int do
      ignore (Misc.intervalle 0 1000)
    done

  include Main(struct let launch () = ignore(loop ()) end)
end

module Prodcons = functor (Dummy : Unit) ->
struct
  (** stock for producer/consumer *)
  type 'a stock = {
    mutable stock : 'a list;
    add : 'a -> unit;
    take : unit -> 'a;
  }
      
  (** create a stock for producer/consumer *)
  let create_stock () =
    let mutex_stock = Mutex.create () in
    let rec this =
      {
	stock = [];
	add =
          (
            fun e ->
              Mutex.lock mutex_stock ;
              this.stock <- e :: this.stock;
              Mutex.unlock mutex_stock ;
          );
	take = 
          (
            let rec f () =
              Mutex.lock mutex_stock ;
              match this.stock with
		| [] ->
                    Mutex.unlock mutex_stock;
                    Thread.yield ();
                    f ()
		| e :: tl ->
                    this.stock <- tl;
                    Mutex.unlock mutex_stock;
                    e
            in
              (fun e -> f ())
          )
      }
    in this
	 
  (** compute for a little while *)
  let random_compute () =
    let rec f r = function 0 -> r
      | n -> f (r*n) (n-1)
    in
      ignore (f 1 (Misc.pseudo_random_int 1000))

  (* create a producer/consumer *)
  let create_pc () =
    let producer, consumer =
      let stock = create_stock () in
	begin
          begin
            fun () ->
              (* produce a lot, else the consumer reaches a dead lock *)
              for i = 0 to 1000 * 10 do
		if Misc.pseudo_random_int 2 = 0 then
                  begin
                    Misc.safe_print_string "add...";
                    stock.add ();
                    Misc.safe_print_string " -> add ok\n";
                  end;
		random_compute ();
              done
          end,
          begin
            fun () ->
              for i = 0 to 1000 do
		if Misc.pseudo_random_int 2 = 0 then
                  begin
                    Misc.safe_print_string "take...";
                    stock.take ();
		    Misc.safe_print_string " -> take ok\n";
                  end;
		random_compute ();
              done
          end
	end
    in
      producer, consumer
	
  let launch () =
    let producer, consumer = create_pc () 
    in
      List.iter
	Thread.join
	(List.map
           (fun e -> Thread.create e ())
           [ producer ; consumer ; ])

  include Main(struct let launch () = ignore(launch()) end)
end




(*  ************************************************************ *)
type numberofthreads = int
type times = int
module type Bench =
    functor (Unit : Unit) ->
      (sig val main : numberofthreads -> unit end)


(* ************************************************************ *)
(* BEGIN Arguments analysis *)
module ArgumentAnalysis =
struct

  let usage = ref (fun () -> ())

  let tryapply f a = try f a with _ -> (!usage () ; exit 42)

  let argv_get i = Array.get Sys.argv i

  let number_of_threads = ref 0

  let do_n_times = ref 0
    
  (* ************************************************************ *)
  (* BEGIN Launch program *)
  module LaunchBench =
    functor (M : Bench) ->
  struct
    let _ = let module Dummy = M(Unit) in
      for i = 1 to !do_n_times do
	Trace.safe_print_string (Printf.sprintf "***Launch #%d***\n" i);
	Dummy.main !number_of_threads
      done;
      Trace.safe_print_string "I've finished my job.\nThe End :-)"
  end
  (* END Launch program *)
  (* ************************************************************ *)

  let benchesdescr = 
    [
      (* name, (descr, action_trigger); *)
      ("prodcons",
       ("\tshared memory",
	fun () ->
	  begin let module M = LaunchBench(Prodcons) in
	    ()
	  end));
    ]

  let availablebenches =
    List.fold_left 
      (
	fun r (name, (descr, _)) ->
	  r ^
	    Printf.sprintf "%s\t%s\n" name descr
      )
      ""
      benchesdescr

  let _  =
    usage := 
      begin fun () ->
	Printf.eprintf "%s name_of_bench number_of_threads do_n_times\nname_of_bench: name of bench\nnumber_of_threads: how many threads you wish\ndo_n_times: how many times you want to run it\n\nThese are the available benches:\n%s" (Sys.argv.(0)) (availablebenches);
	flush stderr
      end
  let _ =
    number_of_threads := tryapply int_of_string (tryapply argv_get 3);
    do_n_times := tryapply int_of_string (tryapply argv_get 2)

  let number_of_threads = !number_of_threads
  let do_n_times = !do_n_times

  let _ =
    let choice = tryapply argv_get 1 in
      (snd (tryapply (List.assoc choice) benchesdescr)) ()

  let usage = !usage
end
(* END Arguments analysis *)
(* ************************************************************ *)
